@@ -1,6 +1,6 @@
 # table.tcl --
 #
-# version 1.8, jeff.hobbs@acm.org
+# version align with tkTable 2.4, jeff.hobbs@acm.org
 # This file defines the default bindings for Tk table widgets
 # and provides procedures that help in implementing those bindings.
 #
@@ -11,6 +11,8 @@
 # afterId -		Token returned by "after" for autoscanning.
 # tablePrev -		The last element to be selected or deselected
 #			during a selection operation.
+# mouseMoved -		Boolean to indicate whether mouse moved while
+#			the button was pressed.
 #--------------------------------------------------------------------------
 
 # tkTableClipboardKeysyms --
@@ -29,11 +31,11 @@ proc tkTableClipboardKeysyms {copy cut paste} {
     bind Table <$paste>	{tk_tablePaste %W}
 }
 
-## Interactive row resizing, affected by -resizeborders option
+## Interactive cell resizing, affected by -resizeborders option
 ##
 bind Table <3>		{
-    ## You might want to check for row returned if you want to
-    ## restrict the resizing of certain rows
+    ## You might want to check for cell returned if you want to
+    ## restrict the resizing of certain cells
     %W border mark %x %y
 }
 bind Table <B3-Motion>	{ %W border dragto %x %y }
@@ -45,10 +47,16 @@ bind Table <1> {
 	tkTableBeginSelect %W [%W index @%x,%y]
 	focus %W
     }
+    array set tkPriv {x %x y %y}
+    set tkPriv(mouseMoved) 0
 }
 bind Table <B1-Motion> {
-    array set tkPriv {x %x y %y}
-    tkTableMotion %W [%W index @%x,%y]
+    # If we already had motion, or we moved more than 1 pixel,
+    # then we start the Motion routine
+    if {$tkPriv(mouseMoved) || abs(%x-$tkPriv(x))>1 || abs(%y-$tkPriv(y))>1} {
+	set tkPriv(mouseMoved) 1
+    }
+    if {$tkPriv(mouseMoved)} { tkTableMotion %W [%W index @%x,%y] }
 }
 bind Table <Double-1> {
     # empty
@@ -73,8 +81,10 @@ bind Table <2> {
     set tkPriv(mouseMoved) 0
 }
 bind Table <B2-Motion> {
-    if {(%x != $tkPriv(x)) || (%y != $tkPriv(y))} { set tkPriv(mouseMoved) 1 }
-    if $tkPriv(mouseMoved) { %W scan dragto %x %y }
+    if {(%x != $tkPriv(x)) || (%y != $tkPriv(y))} {
+	set tkPriv(mouseMoved) 1
+    }
+    if {$tkPriv(mouseMoved)} { %W scan dragto %x %y }
 }
 bind Table <ButtonRelease-2> {
     if {!$tkPriv(mouseMoved)} { tk_tablePaste %W [%W index @%x,%y] }
@@ -88,15 +98,14 @@ if {[string comp {} [info command event]]} {
     tkTableClipboardKeysyms Control-c Control-x Control-v
 }
 
-bind Table <Any-Tab> {
-    # empty to allow Tk focus movement
-}
 # This forces a cell commit if an active cell exists
-# Remove this if you don't want cell commit to occur
-# on every FocusOut
-bind Table <FocusOut> {
+bind Table <<Table_Commit>> {
     catch {%W activate active}
 }
+# Remove this if you don't want cell commit to occur on every
+# Leave of the table.  Another possible choice is <FocusOut>.
+event add <<Table_Commit>> <Leave>
+
 bind Table <Shift-Up>		{tkTableExtendSelect %W -1  0}
 bind Table <Shift-Down>		{tkTableExtendSelect %W  1  0}
 bind Table <Shift-Left>		{tkTableExtendSelect %W  0 -1}
@@ -131,8 +140,8 @@ bind Table <Up>			{tkTableMoveCell %W -1  0}
 bind Table <Down>		{tkTableMoveCell %W  1  0}
 bind Table <Left>		{tkTableMoveCell %W  0 -1}
 bind Table <Right>		{tkTableMoveCell %W  0  1}
-bind Table <Any-KeyPress> {
-    if {[string compare {} %A]} { %W insert active insert %A }
+bind Table <KeyPress> {
+    tkTableInsert %W %A
 }
 bind Table <BackSpace> {
     set tkPriv(junk) [%W icursor]
@@ -145,7 +154,7 @@ bind Table <Escape>		{%W reread}
 
 #bind Table <Return>		{tkTableMoveCell %W 1 0}
 bind Table <Return> {
-    %W insert active insert "\n"
+    tkTableInsert %W "\n"
 }
 
 bind Table <Control-Left>	{%W icursor [expr {[%W icursor]-1}]}
@@ -156,6 +165,35 @@ bind Table <Control-k>		{%W delete active insert end}
 bind Table <Control-equal>	{tkTableChangeWidth %W active  1}
 bind Table <Control-minus>	{tkTableChangeWidth %W active -1}
 
+# Ignore all Alt, Meta, and Control keypresses unless explicitly bound.
+# Otherwise, if a widget binding for one of these is defined, the
+# <KeyPress> class binding will also fire and insert the character,
+# which is wrong.  Ditto for Tab.
+
+bind Table <Alt-KeyPress>	{# nothing}
+bind Table <Meta-KeyPress>	{# nothing}
+bind Table <Control-KeyPress>	{# nothing}
+bind Table <Any-Tab>		{# nothing}
+if {[string match "macintosh" $tcl_platform(platform)]} {
+	bind Table <Command-KeyPress> {# nothing}
+}
+
+# tkTableInsert --
+#
+#   Insert into the active cell
+#
+# Arguments:
+#   w	- the table widget
+#   s	- the string to insert
+# Results:
+#   Returns nothing
+#
+proc tkTableInsert {w s} {
+    if {[string compare $s {}]} {
+	$w insert active insert $s
+    }
+}
+
 # tkTableBeginSelect --
 #
 # This procedure is typically invoked on button-1 presses. It begins
@@ -194,7 +232,7 @@ proc tkTableBeginSelect {w el} {
 		set inc $el
 		set el2 $el
 	    }
-	    if [$w selection includes $inc] {
+	    if {[$w selection includes $inc]} {
 		$w selection clear $el $el2
 	    } else {
 		$w selection set $el $el2
@@ -317,23 +355,41 @@ proc tkTableBeginToggle {w el} {
     if {[string match extended [$w cget -selectmode]]} {
 	set tkPriv(tablePrev) $el
 	$w selection anchor $el
-	if [$w selection includes $el] {
-	    $w selection clear $el
+	if {[$w tag includes title $el]} {
+	    scan $el %d,%d r c
+	    if {$r < [$w cget -titlerows]+[$w cget -roworigin]} {
+		## We're in a column header
+		if {$c < [$w cget -titlecols]+[$w cget -colorigin]} {
+		    ## We're in the topleft title area
+		    set end end
+		} else {
+		    set end [$w index end row],$c
+		}
+	    } else {
+		## We're in a row header
+		set end $r,[$w index end col]
+	    }
 	} else {
-	    $w selection set $el
+	    ## We're in a non-title cell
+	    set end $el
 	}
+	if {[$w selection includes  $end]} {
+	    $w selection clear $el $end
+	} else {
+	    $w selection set   $el $end
+        }
     }
 }
 
 # tkTableAutoScan --
-# This procedure is invoked when the mouse leaves an entry window
+# This procedure is invoked when the mouse leaves an table window
 # with button 1 down. It scrolls the window up, down, left, or
 # right, depending on where the mouse left the window, and reschedules
 # itself as an "after" command so that the window continues to scroll until
 # the mouse moves back into the window or the mouse button is released.
 #
 # Arguments:
-# w - The entry window.
+# w - The table window.
 
 proc tkTableAutoScan {w} {
     global tkPriv
@@ -359,7 +415,8 @@ proc tkTableAutoScan {w} {
 #
 # Moves the location cursor (active element) by the specified number
 # of cells and changes the selection if we're in browse or extended
-# selection mode.
+# selection mode.  If the new cell is "hidden", we skip to the next
+# visible cell if possible, otherwise just abort.
 #
 # Arguments:
 # w - The table widget.
@@ -370,7 +427,24 @@ proc tkTableMoveCell {w x y} {
     global tkPriv
     if {[catch {$w index active row} r]} return
     set c [$w index active col]
-    $w activate [incr r $x],[incr c $y]
+    set cell [$w index [incr r $x],[incr c $y]]
+    while {[string compare [set true [$w hidden $cell]] {}]} {
+	# The cell is in some way hidden
+	if {[string compare $true [$w index active]]} {
+	    # The span cell wasn't the previous cell, so go to that
+	    set cell $true
+	    break
+	}
+	if {$x > 0} {incr r} elseif {$x < 0} {incr r -1}
+	if {$y > 0} {incr c} elseif {$y < 0} {incr c -1}
+	if {[string compare $cell [$w index $r,$c]]} {
+	    set cell [$w index $r,$c]
+	} else {
+	    # We couldn't find a non-hidden cell, just don't move
+	    return
+	}
+    }
+    $w activate $cell
     $w see active
     switch [$w cget -selectmode] {
 	browse {
@@ -422,7 +496,7 @@ proc tkTableDataExtend {w el} {
     if {[string match extended $mode]} {
 	$w activate $el
 	$w see $el
-	if [$w selection includes anchor] {tkTableMotion $w $el}
+	if {[$w selection includes anchor]} {tkTableMotion $w $el}
     } elseif {[string match multiple $mode]} {
 	$w activate $el
 	$w see $el
