@@ -4,7 +4,7 @@
  *	This is the header file for the module that implements
  *	table widgets for the Tk toolkit.
  *
- * Copyright (c) 1997,1998 Jeffrey Hobbs
+ * Copyright (c) 1997-1999 Jeffrey Hobbs
  *
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
@@ -18,49 +18,54 @@
 #include <stdlib.h>
 #include <tk.h>
 #ifdef MAC_TCL
-#include <Xatom.h>
+# include <Xatom.h>
 #else
-#include <X11/Xatom.h>
+# include <X11/Xatom.h>
+#endif /* MAC_TCL */
+
+#if (TCL_MINOR_VERSION == 0) /* Tcl8.0 stuff */
+#define Tcl_GetString(objPtr)	Tcl_GetStringFromObj(objPtr, (int *)NULL)
 #endif
 
-#include "cmd.h"
+/* This EXTERN declaration is needed for Tcl < 8.0.3 */
+#ifndef EXTERN
+# ifdef __cplusplus
+#  define EXTERN extern "C"
+# else
+#  define EXTERN extern
+# endif
+#endif
+
+#ifdef TCL_STORAGE_CLASS
+# undef TCL_STORAGE_CLASS
+#endif
+#ifdef BUILD_tkTable
+# define TCL_STORAGE_CLASS DLLEXPORT
+#else
+# define TCL_STORAGE_CLASS DLLIMPORT
+#endif
+
+#include "mm.h"
 
 #ifdef WIN32
 #   define WIN32_LEAN_AND_MEAN
 #   include <windows.h>
 #   undef WIN32_LEAN_AND_MEAN
-
-/*
- * VC++ has an alternate entry point called DllMain, so we need to rename
- * our entry point.
- */
-
+/* VC++ has an entry point called DllMain instead of DllEntryPoint */
 #   if defined(_MSC_VER)
-#	define EXPORT(a,b) __declspec(dllexport) a b
 #	define DllEntryPoint DllMain
-#   else
-#	if defined(__BORLANDC__)
-#	    define EXPORT(a,b) a _export b
-#	else
-#	    define EXPORT(a,b) a b
-#	endif
 #   endif
+#endif
 
-/* Necessary to get XSync call defined */
-#   include <tkInt.h>
-
-#elif defined(MAC_TCL)
-
-/* Necessary to get XSync call defined */
-#   include <tkInt.h>
-
-#   define EXPORT(a,b) a b
-#else	/* ! WIN32 */
-#   define EXPORT(a,b) a b
-#endif	/* WIN32 */
+#if defined(WIN32) || defined(MAC_TCL)
+/* XSync call defined in the internals for some reason */
+#   ifndef XSync
+#	define XSync(display, bool) {display->request++;}
+#   endif
+#endif /* defn of XSync */
 
 #ifdef INLINE
-#undef INLINE
+#   undef INLINE
 #endif
 #ifdef __GNUC__
 #   define INLINE inline
@@ -68,7 +73,7 @@
 #   define INLINE __inline
 #else
 #   define INLINE
-#endif
+#endif /* INLINE defn */
 
 #ifndef NORMAL_BG
 #   ifdef WIN32
@@ -96,7 +101,7 @@
 #	define HIGHLIGHT	"Black"
 #	define DEF_TABLE_FONT	"Helvetica -12"
 #   endif
-#endif
+#endif /* NORMAL_BG */
 
 #define MAX(A,B)	(((A)>(B))?(A):(B))
 #define MIN(A,B)	(((A)>(B))?(B):(A))
@@ -125,7 +130,7 @@
  * ACTIVE_DISABLED:	Non-zero means the active cell is -state disabled
  * OVER_BORDER:		Non-zero means we are over a table cell border
  * REDRAW_ON_MAP:	Forces a redraw on the unmap
- * NO_SPANS:		prevent cell spans from being used
+ * AVOID_SPANS:		prevent cell spans from being used
  *
  * FIX - consider adding UPDATE_SCROLLBAR a la entry
  */
@@ -142,20 +147,31 @@
 #define ACTIVE_DISABLED		(1L<<10)
 #define OVER_BORDER		(1L<<11)
 #define REDRAW_ON_MAP		(1L<<12)
-#ifdef SPANS
-#define NO_SPANS		(1L<<13)
-#endif
+#define AVOID_SPANS		(1L<<13)
 
 /* Flags for TableInvalidate && TableRedraw */
 #define ROW		(1L<<0)
 #define COL		(1L<<1)
-#define CELL		(ROW|COL)
+#define CELL		(1L<<2)
+
+#define CELL_BAD	(1<<0)
+#define CELL_OK		(1<<1)
+#define CELL_SPAN	(1<<2)
+#define CELL_HIDDEN	(1<<3)
+#define CELL_VIEWABLE	(CELL_OK|CELL_SPAN)
 
 #define INV_FILL	(1L<<3)	/* use for Redraw when the affected
 				 * row/col will affect neighbors */
 #define INV_FORCE	(1L<<4)
 #define INV_HIGHLIGHT	(1L<<5)
 
+/* These alter how the selection set/clear commands behave */
+#define SEL_ROW		(1<<0)
+#define SEL_COL		(1<<1)
+#define SEL_BOTH	(1<<2)
+#define SEL_CELL	(1<<3)
+#define SEL_NONE	(1<<4)
+
 /*
  * Definitions for tablePtr->dataSource, by bit
  */
@@ -174,6 +190,7 @@ typedef enum {
 typedef struct {
   Tk_3DBorder	bg;		/* background color */
   Tk_3DBorder	fg;		/* foreground color */
+  int		bd;		/* cell border width */
   int		relief;		/* relief type */
   Tk_Font	tkfont;		/* Information about text font, or NULL. */
   Tk_Anchor	anchor;		/* default anchor point */
@@ -213,7 +230,6 @@ typedef struct {
 				 * rows when getting selection */
   char *colSep;			/* separator string to place between
 				 * cols when getting selection */
-  int borderWidth;		/* internal borderwidth */
   TableTag defaultTag;		/* the default tag colors/fonts etc */
   char *yScrollCmd;		/* the y-scroll command */
   char *xScrollCmd;		/* the x-scroll command */
@@ -256,6 +272,8 @@ typedef struct {
   int flashTime;		/* The number of ms to flash a cell for */
   int resize;			/* -resizeborders option for interactive
 				 * resizing of borders */
+  int sparse;			/* Whether to use "sparse" arrays by
+				 * deleting empty array elements (default) */
   char *rowTagCmd, *colTagCmd;	/* script to eval for getting row/tag cmd */
   int highlightWidth;		/* Width in pixels of highlight to draw
 				 * around widget when it has the focus.
@@ -298,10 +316,8 @@ typedef struct {
      by the appropriate *Offset factor */
   Tcl_HashTable *colWidths;	/* hash table of non default column widths */
   Tcl_HashTable *rowHeights;	/* hash table of non default row heights */
-#ifdef SPANS
   Tcl_HashTable *spanTbl;	/* table for spans */
   Tcl_HashTable *spanAffTbl;	/* table for cells affected by spans */
-#endif
   Tcl_HashTable *tagTable;	/* table for style tags */
   Tcl_HashTable *winTable;	/* table for embedded windows */
   Tcl_HashTable *rowStyles;	/* table for row styles */
@@ -315,12 +331,22 @@ typedef struct {
 				   for editing the active cell */
   TableTag *activeTagPtr;	/* cache of active composite tag */
   int activeX, activeY;		/* cache offset of active layout in cell */
+
+#ifdef PROCS
+  Tcl_HashTable *inProc;	/* cells where proc is being evaled */
+  int showProcs;		/* whether to show embedded proc (1) or
+				 * its calculated value (0) */
+  int hasProcs;			/* whether table has embedded procs or not */
+#endif
+
   /* The invalid rectangle if there is an update pending */
   int invalidX, invalidY, invalidWidth, invalidHeight;
   int seen[4];			/* see TableUndisplay */
+#ifdef POSTSCRIPT
   /* Pointer to information used for generating Postscript for the canvas.
    * NULL means no Postscript is currently being generated. */
   struct TkPostscriptInfo *psInfoPtr;
+#endif
 } Table;
 
 /*
@@ -343,6 +369,7 @@ typedef struct TableEmbWindow {
   char *create;			/* Script to create window on-demand.
 				 * NULL means no such script.
 				 * Malloc-ed. */
+  int bd;			/* border width for cell around window */
   int relief;			/* relief type */
   int sticky;			/* How to align window in space */
   int padX, padY;		/* Padding to leave around each side
@@ -352,18 +379,71 @@ typedef struct TableEmbWindow {
 				 * recently. */
 } TableEmbWindow;
 
+extern Tk_ConfigSpec tableSpecs[];
+
 extern void	EmbWinDisplay _ANSI_ARGS_((Table *tablePtr, Drawable window,
-					   TableEmbWindow *ewPtr,
-					   TableTag *tagPtr, int x, int y,
-					   int width, int height));
+			TableEmbWindow *ewPtr, TableTag *tagPtr,
+			int x, int y, int width, int height));
 extern void	EmbWinUnmap _ANSI_ARGS_((register Table *tablePtr,
-					 int rlo, int rhi,
-					 int clo, int chi));
+			int rlo, int rhi, int clo, int chi));
 extern void	EmbWinDelete _ANSI_ARGS_((register Table *tablePtr,
-					   TableEmbWindow *ewPtr));
-extern int	TableWindowCmd _ANSI_ARGS_((Table *tablePtr,
-					    Tcl_Interp *interp,
-					    int argc, char *argv[]));
+			TableEmbWindow *ewPtr));
+extern int	Table_WinMove _ANSI_ARGS_((register Table *tablePtr,
+			char *CONST srcPtr, char *CONST destPtr, int flags));
+extern int	Table_WinDelete _ANSI_ARGS_((register Table *tablePtr,
+			char *CONST idxPtr));
+extern int	Table_WindowCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	TableValidateChange _ANSI_ARGS_((Table *tablePtr, int r,
+			int c, char *old, char *new, int index));
+extern void	TableLostSelection _ANSI_ARGS_((ClientData clientData));
+extern void	TableSetActiveIndex _ANSI_ARGS_((register Table *tablePtr));
+
+/*
+ * HEADERS IN TKTABLECMDS
+ */
+
+extern int	Table_ActivateCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_AdjustCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_BboxCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_BorderCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_ClearCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_CurselectionCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_CurvalueCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_GetCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_ScanCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_SeeCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_SelAnchorCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_SelClearCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_SelIncludesCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_SelSetCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_ViewCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+
+/*
+ * HEADERS IN tkTableEdit
+ */
+
+extern int	Table_EditCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern void	TableDeleteChars _ANSI_ARGS_((register Table *tablePtr,
+			int index, int count));
+extern void	TableInsertChars _ANSI_ARGS_((register Table *tablePtr,
+			int index, char *string));
 
 /*
  * HEADERS IN TKTABLETAG
@@ -371,68 +451,83 @@ extern int	TableWindowCmd _ANSI_ARGS_((Table *tablePtr,
 
 extern TableTag *TableNewTag _ANSI_ARGS_((void));
 extern void	TableMergeTag _ANSI_ARGS_((TableTag *baseTag,
-					   TableTag *addTag));
+			TableTag *addTag));
 extern void	TableInvertTag _ANSI_ARGS_((TableTag *baseTag));
 extern void	TableInitTags _ANSI_ARGS_((Table *tablePtr));
 extern TableTag *FindRowColTag _ANSI_ARGS_((Table *tablePtr,
-					     int cell, int type));
+			int cell, int type));
 extern void	TableCleanupTag _ANSI_ARGS_((Table *tablePtr,
-					     TableTag *tagPtr));
-extern int	TableTagCmd _ANSI_ARGS_((Table *tablePtr, Tcl_Interp *interp,
-					 int argc, char **argv));
+			TableTag *tagPtr));
+extern int	Table_TagCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
 
 /*
  * HEADERS IN TKTABLECELL
  */
 
-extern void	TableCellCoords _ANSI_ARGS_((Table *tablePtr, int row,
-					     int col, int *rx, int *ry,
-					     int *rw, int *rh));
+extern int	TableTrueCell _ANSI_ARGS_((Table *tablePtr, int row, int col,
+					   int *trow, int *tcol));
+extern int	TableCellCoords _ANSI_ARGS_((Table *tablePtr, int row,
+			int col, int *rx, int *ry, int *rw, int *rh));
 extern int	TableCellVCoords _ANSI_ARGS_((Table *tablePtr, int row,
-					      int col, int *rx, int *ry,
-					      int *rw, int *rh, int full));
+			int col, int *rx, int *ry,
+			int *rw, int *rh, int full));
 extern void	TableWhatCell _ANSI_ARGS_((register Table *tablePtr,
-					   int x, int y, int *row, int *col));
+			int x, int y, int *row, int *col));
 extern int	TableAtBorder _ANSI_ARGS_((Table *tablePtr, int x, int y,
-					   int *row, int *col));
+			int *row, int *col));
 extern char *	TableGetCellValue _ANSI_ARGS_((Table *tablePtr, int r, int c));
 extern int	TableSetCellValue _ANSI_ARGS_((Table *tablePtr, int r, int c,
-					       char *value));
+			char *value));
 extern char *	TableCellSort _ANSI_ARGS_((Table *tablePtr, char *str));
 extern int	TableGetIcursor _ANSI_ARGS_((Table *tablePtr, char *arg,
-					     int *posn));
-extern int	TableGetIndex _ANSI_ARGS_((register Table *tablePtr, char *str,
-					   int *row_p, int *col_p));
-extern int	TableSetCmd _ANSI_ARGS_((Table *tablePtr, Tcl_Interp *interp,
-					 int argc, char **argv));
-#ifdef SPANS
-extern int	TableSpanCmd _ANSI_ARGS_((Table *tablePtr, Tcl_Interp *interp,
-					  int argc, char **argv));
-#endif SPANS
+			int *posn));
+#define TableGetIcursorObj(tablePtr, objPtr, posnPtr) \
+	TableGetIcursor(tablePtr, Tcl_GetString(objPtr), posnPtr)
+extern int	TableGetIndex _ANSI_ARGS_((register Table *tablePtr,
+			char *str, int *row_p, int *col_p));
+#define TableGetIndexObj(tablePtr, objPtr, rowPtr, colPtr) \
+	TableGetIndex(tablePtr, Tcl_GetString(objPtr), rowPtr, colPtr)
+extern int	Table_SetCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_HiddenCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern int	Table_SpanCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern void	TableSpanSanCheck _ANSI_ARGS_((register Table *tablePtr));
 
 /*
  * HEADERS IN TKTABLEPS
  */
 
-extern int	TablePostscriptCmd _ANSI_ARGS_((Table *tablePtr,
-			Tcl_Interp *interp, int argc, char **argv));
-extern void	Tcl_DStringAppendAll TCL_VARARGS_DEF(Tcl_DString *, arg1);
+#ifdef POSTSCRIPT
+extern int	Table_PostscriptCmd _ANSI_ARGS_((ClientData clientData,
+			Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]));
+extern void	Tcl_DStringAppendAll _ANSI_ARGS_(TCL_VARARGS(Tcl_DString *, arg1));
+#endif
 
 /*
  * HEADERS IN TKTABLE
  */
 
-EXTERN EXPORT(int,Example_Init) _ANSI_ARGS_((Tcl_Interp *interp));
+EXTERN int Tktable_Init		_ANSI_ARGS_((Tcl_Interp *interp));
+EXTERN int Tktable_SafeInit	_ANSI_ARGS_((Tcl_Interp *interp));
 
+int	TableConfigure _ANSI_ARGS_((Tcl_Interp *interp,
+			Table *tablePtr, int objc, Tcl_Obj *CONST objv[],
+			int flags, int forceUpdate));
 extern void	TableGetActiveBuf _ANSI_ARGS_((register Table *tablePtr));
 extern void	ExpandPercents _ANSI_ARGS_((Table *tablePtr, char *before,
 			int r, int c, char *old, char *new, int index,
 			Tcl_DString *dsPtr, int cmdType));
 extern void	TableInvalidate _ANSI_ARGS_((Table *tablePtr, int x, int y,
-					     int width, int height,
-					     int force));
+			int width, int height, int force));
 extern void	TableRefresh _ANSI_ARGS_((register Table *tablePtr,
-					  int arg1, int arg2, int mode));
+			int arg1, int arg2, int mode));
+void	TableGeometryRequest _ANSI_ARGS_((Table *tablePtr));
+void	TableAdjustActive _ANSI_ARGS_((register Table *tablePtr));
+void	TableAdjustParams _ANSI_ARGS_((register Table *tablePtr));
+void	TableConfigCursor _ANSI_ARGS_((register Table *tablePtr));
 
 #define TableInvalidateAll(tablePtr, flags)	\
 	TableInvalidate((tablePtr), 0, 0, Tk_Width((tablePtr)->tkwin),\
@@ -458,5 +553,12 @@ extern void	TableRefresh _ANSI_ARGS_((register Table *tablePtr,
 		      Tk_Height((tablePtr)->tkwin)-(tablePtr)->highlightWidth-1,\
 		      (rowPtr), (colPtr))
 
+/*
+ * end of header
+ * reset TCL_STORAGE_CLASS to DLLIMPORT.
+ */
+#undef TCL_STORAGE_CLASS
+#define TCL_STORAGE_CLASS DLLIMPORT
+
 #endif /* _TKTABLE_H_ */
 
